/*
 * Copyright 2009-2012 Yi-Qing WANG
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file SPLE_lib.cpp
 * @brief library for SPLE denoising algorithm
 * @brief SPLE is implemented mainly with Eigen (Newmat for SVD)
 * @brief PLE is implemented mainly with Newmat
 * @author Yi-Qing WANG <yiqing.wang@polytechnique.edu>
 */


#include "DataProvider.h"
#include "SPLE_lib.h"
#include "io_png.h"
#include <iostream>
#include <omp.h>

using namespace Eigen;
using namespace std;

void vectorArray2Matrix(
		VectorXd ** varray
,		int row_id
,		MatrixXd & vmatrix
){
	int n_cols = vmatrix.cols();
	int n_rows = vmatrix.rows();
	for( int col = 0; col < n_cols; col++ )
		vmatrix.block( 0, col, n_rows, 1 ) = varray[row_id][col];
}

//calc num of patches needed to cover the whole image
//according to a sliding window scheme specified by overlap
//which is the num of columns shared by horizontally neighboring patches
int num_patches(
		int n_pixels
,		int patch_size
,		int overlap
){
	int step = patch_size - overlap;
//	it holds that for some k, n_pixels = patch_size + step * k + something
//	with something = 0 to k-1
	int something = (n_pixels - patch_size) % step;
	int correction;
	if( something == 0 )
		correction = 1;
	else
		correction = 2;
	int k = (n_pixels - something - patch_size)/step;
	return k + correction;
}

//reduce an image into patches
VectorXd ** image2patches(
		MatrixXd const & image
,		int image_rows
,		int image_cols
,		int overlap
,		int patch_size
){
	int map_rows = num_patches( image_rows, patch_size, overlap );
	int map_cols = num_patches( image_cols, patch_size, overlap );
	int data_size = pow( patch_size, 2 );
//	allocate some memory for patches
	VectorXd ** patch_at_coordinates = new VectorXd * [ map_rows ];
	for( int row = 0; row < map_rows; row++ )
		patch_at_coordinates[ row ] = new VectorXd [ map_cols ];
//	the patch upper left corner's coordinates
	int coordinate_j, coordinate_i = -1*patch_size;
//	coordinate cannot exceed max
	int max_coordinate_i = image_rows - patch_size;
	int max_coordinate_j = image_cols - patch_size;
//	sliding window step
	int step = patch_size - overlap;
//	meat
	for( int i = 0; i < map_rows; i++ ){
		coordinate_i = max( 0, min( max_coordinate_i, coordinate_i + step ) );
		coordinate_j = -1*patch_size;
		for( int j = 0; j < map_cols; j++ ){
			coordinate_j = max( 0, min( max_coordinate_j, coordinate_j + step ) );
			MatrixXd patch = image.block( coordinate_i, coordinate_j, patch_size, patch_size ).transpose();
			patch_at_coordinates[i][j] = Map<MatrixXd>( patch.data(), data_size, 1 );
		}
	}
	cout << "INFO: the image is reduced to ( " << map_rows <<  " , " << map_cols << " ) patches." << endl;
	return patch_at_coordinates;
}


//return num of factors given model and standard setting
int ret_num_factors(
		int std_num_factors
,		int data_size
, 		int model
,		int num_models
){
	//remove static if there is model culling
	static int flat_model = num_models - 1;
	static int textural_model = num_models - 2;
	if( model == flat_model )
		return 1;
	else if( model == textural_model )
		return data_size - 1;
	else
		return std_num_factors;
}


//read in the initial config
//the default setting is 18 orientation + 2 mixture
//patch size can be either 8 or 6
void read_config(
		int data_size
,		int std_num_factors
,		int num_models
,		VectorXd *& prior
,		VectorXd *& mus
,		MatrixXd *& factors
 ){
	bool regularize_cov = true;
	if( regularize_cov )
		cout << "INFO: covariance regularization is planned." << endl;
	else
		cout << "INFO: covariance regularization is NOT planned." << endl;;
	int patch_size = (int)sqrt(data_size);
	//initial GMM setup
	DataProvider my_data;
	//mixing weights
	prior = new VectorXd [ 1 ];
	prior[0] = Map<MatrixXd>( my_data.GetPrior( patch_size ), num_models, 1 );
	//model mean
	mus = new VectorXd [ num_models ];
	//model factor loading
	factors = new MatrixXd [ num_models ];
	for( int model = 0; model < num_models; model++ ){
		mus[ model ] = Map<MatrixXd>( my_data.GetMu( patch_size, model ), data_size, 1 );
		MatrixXd cov_mat = Map<MatrixXd>( my_data.GetCov( patch_size, model ), data_size, data_size );
		//all the rest depends on svd
		MatrixXd eig_vecs( data_size, data_size );
		VectorXd eig_vals( data_size );
		deduce_eigens( cov_mat, std_num_factors, data_size, model, num_models, 0, false, false, eig_vecs, eig_vals );
		//determine the number of factors for this model
		int num_factors = ret_num_factors( std_num_factors, data_size, model, num_models );
		//regularize the covariance matrices
		double average_noise_energy = 0;
		int redundancy = data_size - num_factors;
		if( regularize_cov && redundancy > 0 )
			average_noise_energy = eig_vals.tail( redundancy ).sum()/(double)redundancy;
		VectorXd cst;
		cst.setConstant( num_factors, average_noise_energy );
		VectorXd sqrt_eig_vals = (eig_vals.head( num_factors ) - cst).cwiseSqrt();

		//the normalized factors
		factors[ model ] = eig_vecs.block( 0, 0, data_size, num_factors );
		//do not confuse eig vectors with factors
		factors[ model ] *= sqrt_eig_vals.asDiagonal();
		//for testing
		if( model == 0 && false ){
			cout << " model  : " << model << " : " << num_factors << endl;
			//Map<MatrixXd> showp( factors[model].block(0, 0, data_size, 1 ).data(), patch_size, patch_size );
			//cout << factors[model].block(0, 0, data_size, 1 )<< endl;
		}
	}
}


//deduce factors and their eigen-values from a covariance matrix
//this routine is needed as several places in this implementation
void deduce_eigens(
		MatrixXd & cov_mat
,		int std_num_factors
,		int data_size
,		int model
,		int num_models
, 		double squared_sigma
,		bool inverse_eig_vals
,		bool trim
,		MatrixXd & eig_vecs
,		VectorXd & eig_vals
){
		JacobiSVD<MatrixXd> svd(cov_mat, ComputeThinU);
		MatrixXd U_matrix = svd.matrixU();
		VectorXd D_vector = svd.singularValues();
		//determine the number of factors for this model
		if( trim ){
			int num_factors = ret_num_factors( std_num_factors, data_size, model, num_models );
			eig_vecs = U_matrix.block( 0, 0, data_size, num_factors ).transpose();
			if( inverse_eig_vals ){
				VectorXd ss;
				ss.setConstant( num_factors, squared_sigma );
				ss += D_vector.head( num_factors );
				eig_vals = ss.cwiseInverse();
			}else
				eig_vals = D_vector.head( num_factors );
		}else{
			eig_vals = D_vector;
			eig_vecs = U_matrix;
		}
}

//one of the most time consuming part of the algorithm
//calc for all (patch, model) couple their probability (scaled)
//and form the patch map for SURE-aided adaptive filtering
void calc_responsibilities(
		VectorXd ** noisy_patches_at_coordinates
, 		MatrixXd * factors
, 		VectorXd * mus
,		VectorXd * prior
,		int map_rows
,		int map_cols
,		int data_size
,		int std_num_factors
,	 	double sigma
, 		VectorXd **& responsibility
, 		MatrixXi & patch_map
){

	cout << "INFO: calculating model responsibilities..." << endl;
//	allocate memory just once for later responsibility will not be NULL any more
	if( responsibility == NULL ){
		responsibility = new VectorXd * [ map_rows ];
		for( int row = 0; row < map_rows; row++ )
			responsibility[ row ] = new VectorXd [ map_cols ];
	}

//	compute the models' eigenvalues and eigenvectors to avoid repeated operations
	int num_models = prior[0].rows();
//	inverse eigenvalues for density calculation
	VectorXd * inv_eig_vals = new VectorXd [ num_models ];
//	eigenvectors
	MatrixXd * eig_vecs = new MatrixXd[ num_models ];

	double squared_sigma = sigma * sigma;

//	the first set of factors are orthogonal while it might not be the case in later iterations
//	as EM is supposed to approximate, rather than solve the density maximization problem
//	this part is taken out because the values in it will be repeatedly used later in the function
	#pragma omp parallel for schedule( static )
	for( int model = 0; model < num_models; model++ ){
		MatrixXd cov = factors[ model ] * factors[ model ].transpose();
		int num_factors = ret_num_factors( std_num_factors, data_size, model, num_models );
		MatrixXd temp_eig_vecs( num_factors, data_size );
		VectorXd temp_eig_vals( num_factors );
		deduce_eigens( cov, std_num_factors, data_size, model, num_models, squared_sigma, true, true, temp_eig_vecs, temp_eig_vals );
		eig_vecs[ model ] = temp_eig_vecs;
		inv_eig_vals[ model ] = temp_eig_vals;
	}

//	now, we can compute the patch responsibilities, up to a multiplicative constant
//	if for one particular patch, no model yields a positive density, let us default
//	the unknown model to the textural_model at the risk of producing artifacts
//	or to the flat one at the risk of producing equally undesirable blur
	int flat_model = num_models - 1;
	int textural_model = num_models - 2;
	int default_model = flat_model;
	VectorXd fake;
	fake.setZero( num_models );
	fake( default_model ) = 1;


//	get the business started here: calc responsibility as well as the patch map
	#pragma omp parallel for schedule( static )
	for( int row_id = 0; row_id < map_rows; row_id++ )
		for( int col_id = 0; col_id < map_cols; col_id++ ){
			double highest;
			int best_model;
			VectorXd densities( num_models );
			for( int model = 0; model < num_models; model++ ){
				int num_factors = ret_num_factors( std_num_factors, data_size, model, num_models );
				double val = prior[0](model)*gaussian_density_EM( noisy_patches_at_coordinates[row_id][col_id], mus[model], inv_eig_vals[model], eig_vecs[model], squared_sigma, num_factors);
				densities( model ) = val;
				if( model == 0 || val > highest ){
					highest = val;
					best_model = model;
				}
			}
			double test = densities.sum();
			if( test == 0 || isinf(test) ){
				//only report serious error
				if( isinf(test) ){
					cerr << "#######################################################################" << endl;
					cerr << "Error: wrong density sum : " << test << " The program could break"  << endl;
					cerr << "problem patch: row_id : " << row_id << " col_id : " << col_id << endl;
					cerr << "#######################################################################" << endl;
				}
				responsibility[ row_id ][ col_id ] = fake;
				patch_map( row_id, col_id ) = default_model;
			}else{
				//normalize the density so they become MAP probability
				responsibility[ row_id ][ col_id ] = densities/test;
				patch_map( row_id, col_id ) = best_model;
			}
		}

	delete [] eig_vecs;
	delete [] inv_eig_vals;
}


//calc gaussian density for EM up to a multiplicative factor
//where x is the variable that samples the density
//mu is the gaussian vector's expectation
double gaussian_density_EM(
			VectorXd const & x
,			VectorXd const & mu
, 			VectorXd const & inv_eig_vals
,			MatrixXd const & eig_vecs
,			double squared_sigma
,			int num_factors
){
	//projection
	VectorXd centered = x - mu;
	VectorXd coordinates = eig_vecs * centered;
	//logNumerator
	double logNumerator = coordinates.cwiseAbs2().dot(inv_eig_vals)+(centered.squaredNorm()-coordinates.squaredNorm())/squared_sigma;
	//static because the density is only computed for the first channel, otherwise should be removed
	static double logSSigma = log(squared_sigma);
	static int data_size = x.rows();
	//300 helps ensure that density values remain in an acceptable numerical range
	double result = exp(300 + 0.5*(logSSigma*(num_factors - data_size) - logNumerator));
	//remaining terms
	double inv_product = 1e200;
	for( int i = 0; i < num_factors; i++ )
		inv_product *= inv_eig_vals(i);
	//ok
	result *= sqrt(inv_product);
	if( isinf(result) )
		cerr << "ERROR: find inf density. Please adjust the multiplicative constants" << endl;
	return result;
}

//update the probabilistic PCA mixture with the two stage GEM
void update_PPCA(
		VectorXd ** noisy_patches
,		VectorXd ** responsibility
,		VectorXd * prior
,		int num_models
,		int map_rows
,		int map_cols
,		double sigma
,		VectorXd *& mus
,		MatrixXd *& factors
){
	cout << "INFO: updating the mixture ..." << endl;
	int data_size = noisy_patches[0][0].rows();


//	update the prior probability and the model expectations at the same time
//	updating rule is very very intuitive, basically it says that given a mixture model
//	you can never be sure which model a patch belongs to, no matter where you find it
//	because a Gaussian multivariate distribution has an infinite support

	MatrixXd updated_means;
	updated_means.setZero(data_size, num_models);
	VectorXd total_weights;
	total_weights.setZero(num_models);
	for( int row = 0; row < map_rows; row++ ){
		MatrixXd patch_matrix( data_size, map_cols );
		MatrixXd resp_matrix( num_models, map_cols );
		vectorArray2Matrix( noisy_patches, row, patch_matrix );
		vectorArray2Matrix( responsibility, row, resp_matrix );
		updated_means += patch_matrix * resp_matrix.transpose();
		total_weights += resp_matrix.rowwise().sum();
	}
	updated_means *= total_weights.cwiseInverse().asDiagonal();
	double num_patches = map_rows * map_cols;
	prior[0] = total_weights/num_patches;


//	now update the factors, memory allocation first
//	OmegaF refers to the product of Omega (estimated model covariance matrix up to a multiplicative constant) and Factors (see doc)
	MatrixXd * OmegaFs = new MatrixXd [num_models];
	MatrixXd * centers = new MatrixXd [num_models];
	int std_num_factors = factors[0].cols();
//	#pragma omp parallel for
	for( int model = 0; model < num_models; model++ ){
		mus[model] = updated_means.block(0, model, data_size, 1);
		centers[model] = mus[model].rowwise().replicate(map_cols);
		int num_factors = ret_num_factors( std_num_factors, data_size, model, num_models );
		OmegaFs[model].setZero(data_size, num_factors);
	}
//	continue. first calculate OmegaFs
//	it is the next chunk that consumes the most time in this routine
	#pragma omp parallel for schedule( static )
	for( int row = 0; row < map_rows; row++ ){
		MatrixXd patch_matrix( data_size, map_cols );
		MatrixXd resp_matrix( num_models, map_cols );
		vectorArray2Matrix( noisy_patches, row, patch_matrix );
		vectorArray2Matrix( responsibility, row, resp_matrix );
		for( int model = 0; model < num_models; model++ ){
			MatrixXd centered = patch_matrix - centers[model];
			MatrixXd tmp = centered*(resp_matrix.row(model).asDiagonal() * (factors[model].transpose() * centered).transpose());
			#pragma omp critical
			{
				OmegaFs[model] += tmp;
			}
		}
	}
//	release
	delete [] centers;


//	the final two linear equations to solve
	double squared_sigma = sigma * sigma;
	#pragma omp parallel for schedule( static )
	for( int model = 0; model < num_models; model++ ){
//		to understand the rest, please refer to the doc
		int num_factors = ret_num_factors( std_num_factors, data_size, model, num_models );
		MatrixXd identity;
		identity.setIdentity(num_factors, num_factors);
		MatrixXd M = factors[model].transpose()*factors[model] + identity*squared_sigma;
//		note that OmegaFs haven't been normalized
		double normalizer = squared_sigma * prior[0](model) * num_patches;
		MatrixXd intermediate = M.ldlt().solve(factors[model].transpose())*OmegaFs[model] + identity*normalizer;
		factors[model] = intermediate.transpose().colPivHouseholderQr().solve(OmegaFs[model].transpose()).transpose();
	}
//	release
	delete [] OmegaFs;
}


//write image and release memory
void imwrite(
		const char * file_name
, 		MatrixXd * image
, 		int num_channels
){
	int image_cols = image[0].cols();
	int image_rows = image[0].rows();
	int pixels_per_channel = image_cols * image_rows;
	float * output = new float [ pixels_per_channel * num_channels ];
	//this part should be straightforward but still be careful with the order
	#pragma omp parallel for schedule( static )
	for(int j = 0; j < image_cols; j++)
		for(int i = 0; i < image_rows; i++)
			for(int ch = 0; ch < num_channels; ch++)
				output[ ch*pixels_per_channel + i*image_cols + j ] = (float) image[ ch ](i,j);
	write_png_f32( file_name, output, (size_t) image_cols, (size_t) image_rows, (size_t) num_channels );
	delete [] output;
	cout << "INFO: write the image " << file_name << " to local folder." << endl;
}

//display the colorful patch_map
void print_patch_map(
		MatrixXi const & patch_map
,		bool is_patch_map
){
	//read in model to color mapping
	DataProvider mydata;
	double * cm = mydata.GetColormap();
	//in total 64 different colors
	Map<MatrixXd> colormap( cm, 64, 3 );


	//allocate memory for the patch_map
	int map_rows = patch_map.rows();
	int map_cols = patch_map.cols();
	MatrixXd temp( map_rows, map_cols );
	MatrixXd * patch_map_ptr = new MatrixXd [3];
	for( int ch = 0; ch < 3; ch++ )
		patch_map_ptr[ ch ] = temp;


	//range allows to print out other maps
	double range = 20;
	if( !is_patch_map )
		range = patch_map.maxCoeff() - patch_map.minCoeff() + 1;
	//fill in
	#pragma omp parallel for schedule( static )
	for( int row = 0; row < map_rows; row++ )
		for( int col = 0; col < map_cols; col++ ){
			int scale = patch_map( row, col );
			//+1 because scale ranges from 0 to 19
			scale = ceil( 64.0/range*(scale+1) ) - 1;
			for( int ch = 0; ch < 3; ch++ )
				patch_map_ptr[ch](row, col) = colormap(scale, ch)*255;
		}


	//print it out
	if( is_patch_map )
		imwrite( "patchmap.png", patch_map_ptr, 3 );
	else
		imwrite( "othermap.png", patch_map_ptr, 3 );
}
